@@ -8,6 +8,7 @@ from pandas import json_normalize
 from pandas import Series
 from pandas import to_datetime
 from py42 import exceptions
+from py42.clients.settings.device_settings import IncydrDeviceSettings
 from py42.exceptions import Py42NotFoundError
 
 from code42cli.bulk import generate_template_cmd_factory
@@ -35,7 +36,13 @@ def devices(state):
 
 
 device_guid_argument = click.argument(
-    "device-guid", type=str, callback=lambda ctx, param, arg: _verify_guid_type(arg),
+    "device-guid",
+    type=str,
+    callback=lambda ctx, param, arg: _verify_guid_type(arg),
+)
+
+new_device_name_option = click.option(
+    "-n", "--new-device-name", help="The new name for the device.", required=True
 )
 
 
@@ -60,6 +67,15 @@ purge_date_option = click.option(
 )
 
 
+@devices.command()
+@device_guid_argument
+@new_device_name_option
+@sdk_options()
+def rename(state, device_guid, new_device_name):
+    """Rename a device with Code42. Requires the device GUID to rename."""
+    _change_device_name(state.sdk, device_guid, new_device_name)
+
+
 @devices.command()
 @device_guid_argument
 @change_device_name_option(
@@ -144,9 +160,24 @@ def _update_cold_storage_purge_date(sdk, guid, purge_date):
 
 
 def _change_device_name(sdk, guid, name):
-    device_settings = sdk.devices.get_settings(guid)
-    device_settings.name = name
-    sdk.devices.update_settings(device_settings)
+    try:
+        device_settings = sdk.devices.get_settings(guid)
+        if isinstance(device_settings, IncydrDeviceSettings):
+            raise Code42CLIError(
+                "Failed to rename device. Incydr devices cannot be renamed."
+            )
+        device_settings.name = name
+        sdk.devices.update_settings(device_settings)
+    except KeyError:
+        raise Code42CLIError(
+            "Failed to rename device. This device is missing expected settings fields."
+        )
+    except exceptions.Py42ForbiddenError:
+        raise Code42CLIError(
+            f"You don't have the necessary permissions to rename the device with GUID '{guid}'."
+        )
+    except exceptions.Py42NotFoundError:
+        raise Code42CLIError(f"The device with GUID '{guid}' was not found.")
 
 
 @devices.command()
@@ -357,11 +388,8 @@ def list_devices(
         df = _add_usernames_to_device_dataframe(state.sdk, df)
     if include_legal_hold_membership:
         df = _add_legal_hold_membership_to_device_dataframe(state.sdk, df)
-    if df.empty:
-        click.echo("No results found.")
-    else:
-        formatter = DataFrameOutputFormatter(format)
-        formatter.echo_formatted_dataframe(df)
+    formatter = DataFrameOutputFormatter(format)
+    formatter.echo_formatted_dataframes(df)
 
 
 def _add_legal_hold_membership_to_device_dataframe(sdk, df):
@@ -483,7 +511,12 @@ def _break_backup_usage_into_total_storage(backup_usage):
 @format_option
 @sdk_options()
 def list_backup_sets(
-    state, active, inactive, org_uid, include_usernames, format,
+    state,
+    active,
+    inactive,
+    org_uid,
+    include_usernames,
+    format,
 ):
     """Get information about many devices and their backup sets."""
     if inactive:
@@ -493,11 +526,8 @@ def list_backup_sets(
     if include_usernames:
         df = _add_usernames_to_device_dataframe(state.sdk, df)
     df = _add_backup_set_settings_to_dataframe(state.sdk, df)
-    if df.empty:
-        click.echo("No results found.")
-    else:
-        formatter = DataFrameOutputFormatter(format)
-        formatter.echo_formatted_dataframe(df)
+    formatter = DataFrameOutputFormatter(format)
+    formatter.echo_formatted_dataframes(df)
 
 
 def _add_backup_set_settings_to_dataframe(sdk, devices_dataframe):
@@ -523,9 +553,9 @@ def _add_backup_set_settings_to_dataframe(sdk, devices_dataframe):
                     "destinations": [
                         destination for destination in backup_set.destinations.values()
                     ],
-                    "included files": backup_set.included_files,
-                    "excluded files": backup_set.excluded_files,
-                    "filename exclusions": backup_set.filename_exclusions,
+                    "included files": list(backup_set.included_files),
+                    "excluded files": list(backup_set.excluded_files),
+                    "filename exclusions": list(backup_set.filename_exclusions),
                     "locked": backup_set.locked,
                 }
                 for backup_set in current_device_settings.backup_sets
@@ -550,13 +580,14 @@ def bulk(state):
 
 
 _bulk_device_activation_headers = ["guid"]
-
+_bulk_device_rename_headers = ["guid", "name"]
 
 devices_generate_template = generate_template_cmd_factory(
     group_name="devices",
     commands_dict={
         "reactivate": _bulk_device_activation_headers,
         "deactivate": _bulk_device_activation_headers,
+        "rename": _bulk_device_rename_headers,
     },
     help_message="Generate the CSV template needed for bulk device commands.",
 )
@@ -638,3 +669,37 @@ def bulk_reactivate(state, csv_rows, format):
         raise_global_error=False,
     )
     formatter.echo_formatted_list(result_rows)
+
+
+@bulk.command(name="rename")
+@read_csv_arg(headers=_bulk_device_rename_headers)
+@format_option
+@sdk_options()
+def bulk_rename(state, csv_rows, format):
+    """Rename all devices from the provided CSV containing a 'guid' and a 'name' column."""
+
+    # Initialize the SDK before starting any bulk processes
+    # to prevent multiple instances and having to enter 2fa multiple times.
+    sdk = state.sdk
+
+    csv_rows[0]["renamed"] = "False"
+    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
+    stats = create_worker_stats(len(csv_rows))
+
+    def handle_row(**row):
+        try:
+            _change_device_name(sdk, row["guid"], row["name"])
+            row["renamed"] = "True"
+        except Exception as err:
+            row["renamed"] = f"False: {err}"
+            stats.increment_total_errors()
+        return row
+
+    result_rows = run_bulk_process(
+        handle_row,
+        csv_rows,
+        progress_label="Renaming devices:",
+        stats=stats,
+        raise_global_error=False,
+    )
+    formatter.echo_formatted_list(result_rows)
